# -*- coding: utf-8 -*-

# This code is a part of sunshine package: https://github.com/letuananh/sunshine
# :copyright: (c) 2013-2021 Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

from datetime import datetime
import logging
import json
import uuid
from collections.abc import Sequence, Mapping

from django.conf import settings
from django.utils.timezone import make_aware
from django.http import HttpResponse
try:
    from django.contrib.admin.models import LogEntry
    from django.contrib.admin.models import ADDITION, DELETION, CHANGE
    from django.contrib.contenttypes.models import ContentType
    _ADMIN_MODEL_AVAILABLE = True
except Exception:
    _ADMIN_MODEL_AVAILABLE = False
from django.apps import apps
from django.urls import include, path


TIME_ZONE_DST = settings.TIME_ZONE_DST


def datetime_now():
    ''' Generate aware datetime object of current time (now) '''
    return make_aware(datetime.now(), is_dst=TIME_ZONE_DST)


def read_settings(key, default_value=''):
    return getattr(settings, key, default_value)


def log_action(request, instance, **kwargs):
    if _ADMIN_MODEL_AVAILABLE:
        LogEntry.objects.log_action(
            user_id=request.user.id,
            content_type_id=ContentType.objects.get_for_model(instance).pk,
            object_id=instance.id,
            object_repr=repr(instance),
            **kwargs)


def log_add(request, instance, **kwargs):
    log_action(request, instance, action_flag=ADDITION, **kwargs)


def log_update(request, instance, **kwargs):
    log_action(request, instance, action_flag=CHANGE, **kwargs)


def log_delete(request, instance, **kwargs):
    log_action(request, instance, action_flag=DELETION, **kwargs)


def jsonp(func):
    ''' JSON/JSONP decorator '''
    def decorator(request, *args, **kwargs):
        objects = func(request, *args, **kwargs)
        # ignore HttpResponse
        if isinstance(objects, HttpResponse):
            return objects
        # JSON/JSONP response
        data = json.dumps(objects)
        if 'callback' in request.GET:
            callback = request.GET['callback']
        elif 'callback' in request.POST:
            callback = request.POST['callback']
        else:
            return HttpResponse(data, "application/json")
        # is JSONP
        # logging.debug("A jsonp response")
        data = '{c}({d});'.format(c=callback, d=data)
        return HttpResponse(data, "application/javascript")
    return decorator


def random_filename(ext=''):
    ''' Generate a random file name '''
    if ext:
        return str(uuid.uuid4()) + ext
    else:
        return str(uuid.uuid4())


def find_module_info():
    modules = []
    for app in apps.get_app_configs():
        app_path = getattr(app, "sunshine_path", None)
        if app_path:
            app_url = f"{app.module.__name__}.urls"
            logging.getLogger(__name__).info(f"Loading sunshine module [{app}] | URL mapping: {repr(app_path)} ==> {repr(app_url)}")
            modules.append(app)
    return modules


def find_modules():
    potentials = []
    for app in apps.get_app_configs():
        app_path = getattr(app, "sunshine_path", None)
        if app_path:
            app_url = f"{app.module.__name__}.urls"
            app_ns = getattr(app, "namespace", None)
            logging.getLogger(__name__).info(f"Loading sunshine module [{app}] | URL mapping: {repr(app_path)} ==> {repr(app_url)}")
            if app_ns:
                potentials.append(path(app_path, include(app_url, namespace=app_ns)))
            else:
                potentials.append(path(app_path, include(app_url)))
    return potentials


def find_menuitems():
    menu_items = []
    for app in apps.get_app_configs():
        app_menu = getattr(app, "sunshine_menu", None)
        if app_menu:
            for item in app_menu:
                if isinstance(item, Sequence):
                    menu_items.append(Menu(*item))
                elif isinstance(item, Mapping):
                    menu_items.append(Menu(**item))
                else:
                    raise ValueError(f"Unknown menu mapping found ({item})")
    return menu_items


def build_menu(request=None):
    items = find_menuitems()
    menu = {'name': '', 'items': [], 'children': {}}
    uid = 1
    for idx, item in enumerate(items, start=1):
        if request is not None:
            item.request = request
            if not item.has_access:
                continue
        current = menu
        if item.groups:
            for group in item.groups:
                current = current['children'].setdefault(group, {'name': group, 'items': [], 'children': {}, 'div_id': uid})
                uid += 1
        # add this menu item to this group
        current['items'].append(item)
    return menu


class Menu:
    def __init__(self, name, verify=None, title='', icon='', groups=None, type='', div_id=None):
        self.__name = name
        self.__verify = verify
        self.__request = None
        self.__title = title
        self.__icon = icon
        self.__menu_type = '' if not type else str(type).lower()
        self.__groups = tuple(groups) if groups else {}
        self.__div_id = div_id

    @property
    def name(self):
        return self.__name

    @property
    def groups(self):
        return self.__groups

    @property
    def title(self):
        return self.__title if self.__title else self.name

    @property
    def menu_type(self):
        return self.__menu_type

    @property
    def icon(self):
        return self.__icon if self.__icon else 'fa-cog'

    @property
    def verify(self):
        return self.__verify if self.__verify else lambda x: True

    @property
    def has_access(self):
        return self.verify is not None and self.verify(self.request)

    @property
    def request(self):
        return self.__request

    @request.setter
    def request(self, request):
        self.__request = request

    def __repr__(self):
        menu_type = self.menu_type + '::' if self.menu_type else ''
        if self.name != self.title:
            return f"<{menu_type}{self.name}>[{self.title}]"
        else:
            return f"<{menu_type}{self.name}>"

    def __str__(self):
        return repr(self)



# -*- coding: utf-8 -*-

# This code is a part of sunshine package: https://github.com/letuananh/sunshine
# :copyright: (c) 2013-2021 Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

from django.apps import AppConfig
from django.conf import settings

SUNSHINE_ADMIN_GROUPS = getattr(settings, 'SUNSHINE_ADMIN_GROUPS', [])


def is_developer(user):
    return user.is_superuser or \
        (SUNSHINE_ADMIN_GROUPS and
         user.groups.filter(name__in=SUNSHINE_ADMIN_GROUPS).exists()
         )


def is_developer_request(req):
    try:
        return is_developer(req.user)
    except Exception:
        return False


class SunshinePortalConfig(AppConfig):
    name = 'sunshine'
    verbose_name = 'Sunshine Portal'
    sunshine_menu = [
        {'title': 'System Dashboard',
         'name': 'system_dashboard',
         'verify': is_developer_request,
         'groups': ['Developer']
         },
        {'name': 'admin:index',
         'verify': is_developer_request,
         'title': 'Django Admin',
         'icon': 'fa-tools',
         'groups': ['Developer']}
    ]



# -*- coding: utf-8 -*-

# This code is a part of sunshine package: https://github.com/letuananh/sunshine
# :copyright: (c) 2013-2021 Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

from django.conf import settings
from .apps import is_developer
from .apps import is_developer_request
from .common import read_settings, build_menu


SUNSHINE_ADMIN_GROUPS = getattr(settings, 'SUNSHINE_ADMIN_GROUPS', [])
TIME_ZONE_DST = settings.TIME_ZONE_DST
DEBUG = settings.DEBUG
SUNSHINE_ENV = read_settings('SUNSHINE_ENV', 'DEV')
SUNSHINE_BRAND = read_settings('SUNSHINE_BRAND', f"{SUNSHINE_ENV} Sunshine <sup>2</sup>")
DEFAULT_LOGO = 'sunshine/images/local.png'
SUNSHINE_DEFAULT_TITLE = read_settings('SUNSHINE_DEFAULT_TITLE', 'Sunshine Portal')
LOGO_MAP = {
    'DEV': 'sunshine/images/local.png',
    'UAT': 'sunshine/images/uat.png',
    'PROD': 'sunshine/images/prod.png'
}
SUNSHINE_ADMIN_LOGO = read_settings('SUNSHINE_ADMIN_LOGO',
                                    LOGO_MAP[SUNSHINE_ENV] if SUNSHINE_ENV in LOGO_MAP else DEFAULT_LOGO)

COPYRIGHT_FOOTER = read_settings(
    "SUNSHINE_COPYRIGHT",
    """Sunshine Portal, &copy; <a target='_blank' href='https://github.com/letuananh/sunshine'>Le Tuan Anh</a>, 2013""")


def username(request):
    ''' Get current user's login account name '''
    if not hasattr(request, 'user'):
        username = 'Anonymous'
    elif request.user.is_authenticated:
        # try to user fullname first ...
        username = "{} {}".format(request.user.first_name, request.user.last_name).strip()
        if not username:
            username = request.user.username
    else:
        username = 'Anonymous'
    return username


def setup_sunshine_request(request):
    menu = build_menu(request)
    sidebar_toggle = request.session.get('sidebar_toggle', True) if hasattr(request, 'session') else True
    return {
        # sunshine site configuration
        'SUNSHINE_BRAND': SUNSHINE_BRAND,
        'SUNSHINE_ADMIN_LOGO': SUNSHINE_ADMIN_LOGO,
        'SUNSHINE_COPYRIGHT': COPYRIGHT_FOOTER,
        'SUNSHINE_ENV': SUNSHINE_ENV,
        'DEBUG': DEBUG,
        'SUNSHINE_DEFAULT_TITLE': SUNSHINE_DEFAULT_TITLE,
        # sunshine menu
        'sunshine_main_menu': menu,
        # user related info
        'username': username(request),
        'access_dev': is_developer_request(request),
        'sidebar_toggle': sidebar_toggle
    }



# -*- coding: utf-8 -*-

# This code is a part of django-sunshine package: https://github.com/letuananh/sunshine
# :copyright: (c) 2013-2021 Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

__author__ = "Le Tuan Anh"
__email__ = "tuananh.ke@gmail.com"
__copyright__ = "Copyright (c) 2013-2021, Le Tuan Anh <tuananh.ke@gmail.com>"
__credits__ = []
__license__ = "MIT License"
__description__ = "A bootstrap-based boilerplate for building modern Django websites"
__url__ = "https://github.com/letuananh/sunshine"
__issue__ = "https://github.com/letuananh/sunshine/issues/"
__keywords__ = ["bootstrap", "django", "portal"]
__maintainer__ = "Le Tuan Anh"
__version_major__ = "0.4.1"  # follow PEP-0440
__version__ = "{}a1".format(__version_major__)
__version_long__ = "{} - Alpha 1".format(__version_major__)
__status__ = "3 - Alpha"
# __status__ = "5 - Production/Stable"



# -*- coding: utf-8 -*-

# This code is a part of sunshine package: https://github.com/letuananh/sunshine
# :copyright: (c) 2013-2021 Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

import os
import sys
import logging
from pathlib import Path

from django import get_version as django_version
from django.urls import reverse
from django.conf import settings
from django.shortcuts import render, redirect
from django.http import FileResponse
from django.contrib.auth import logout as django_logout
from django.views.decorators.csrf import csrf_protect
from django.contrib.auth.decorators import login_required
from django.contrib.auth.decorators import user_passes_test

import humanize

from .common import jsonp
from .common import find_module_info
from .context_processors import is_developer
from .__version__ import __version__

MEDIA_ROOT = Path(settings.MEDIA_ROOT)  # also for serving CDN
SUNSHINE_CDN_PERMISSIONS = getattr(settings, "SUNSHINE_CDN_PERMISSIONS", {})
SUNSHINE_TITLE = getattr(settings, "SUNSHINE_DEFAULT_TITLE", "Sunshine Portal Index")
SUNSHINE_PUBLIC_WELCOME = getattr(settings, "SUNSHINE_PUBLIC_WELCOME", "")


SUNSHINE_HOME = getattr(settings, 'SUNSHINE_HOME', '')
LOGIN_URL = getattr(settings, 'LOGIN_URL', '/login')


def index(request):
    ''' Default entry point for the public '''
    if hasattr(request, 'user') and request.user.is_authenticated and SUNSHINE_HOME:
        return redirect(SUNSHINE_HOME)
    else:
        context = {'title': SUNSHINE_TITLE,
                   'SUNSHINE_PUBLIC_WELCOME': SUNSHINE_PUBLIC_WELCOME,
                   'version': __version__}
        return render(request, 'sunshine/public.html', context)


def public(request):
    context = {'title': SUNSHINE_TITLE,
               'SUNSHINE_PUBLIC_WELCOME': SUNSHINE_PUBLIC_WELCOME,
               'version': __version__}
    return render(request, 'sunshine/public.html', context)


@login_required()
def home(request):
    return render(request, 'sunshine/home.html',
                  {'version': __version__})


@login_required()
def profile(request):
    return render(request, 'sunshine/profile.html',
                  {'version': __version__})
    

def logout(request):
    django_logout(request)
    return redirect(LOGIN_URL)


# ------------------------------------------------------------------------------
# Alerts
# ------------------------------------------------------------------------------

@jsonp
@csrf_protect
@login_required()
def api_fetch_notifications(request):
    return {'alerts': [], 'inbox': []}
    # return notifications


# ------------------------------------------------------------------------------
# Sidebar
# ------------------------------------------------------------------------------

@jsonp
@csrf_protect
@login_required()
def api_sidebar_toggle(request, status=0):
    try:
        request.session['sidebar_toggle'] = (status != 0)
        return {'status': True, 'state': request.session.get('sidebar_toggle', False)}
    except Exception:
        logging.getLogger(__name__).exception("Sidebar API error")
        return {'status': False, 'state': request.session.get('sidebar_toggle', False)}


# ------------------------------------------------------------------------------
# System tools
# ------------------------------------------------------------------------------

@login_required()
@csrf_protect
@user_passes_test(is_developer)
def system_dashboard(request):
    context = {"sunshine_version": __version__,
               'django_version': django_version(),
               'python_version': sys.version}
    # find modules
    modules = []
    for m in find_module_info():
        name = getattr(m, 'verbose_name', '')
        if not name:
            name = getattr(m, 'name')
        version = getattr(m, 'version', 'N/A')
        build = getattr(m, 'version_build', '')
        modules.append({'name': name,
                        'version': version,
                        'build': build})
    context['modules'] = modules
    # show free space
    try:
        statvfs = os.statvfs(MEDIA_ROOT)
        context['space_used'] = humanize.naturalsize(statvfs.f_frsize * statvfs.f_blocks)     # Size of filesystem in bytes
        context['space_free'] = humanize.naturalsize(statvfs.f_frsize * statvfs.f_bfree)      # Actual number of free bytes
        context['space_usable'] = humanize.naturalsize(statvfs.f_frsize * statvfs.f_bavail)
        # Number of free bytes that ordinary users
        # are allowed to use (excl. reserved space)
    finally:
        pass
    return render(request, 'sunshine/system_dashboard.html', context)


# ------------------------------------------------------------------------------
# Basic built-in CDN
# ------------------------------------------------------------------------------

def cdn_home(request):
    return render(request, 'sunshine/cdn_index.html')


@login_required()
def cdn_serve_avatar(request, username):
    context = {'module': 'avatars',
               'res_id': username}
    options = [MEDIA_ROOT / "avatars" / username,
               MEDIA_ROOT / "avatars" / f"{username}.jpg",
               MEDIA_ROOT / "avatars" / f"{username}.png",
               MEDIA_ROOT / "avatars" / "_default_.png"]
    for f in options:
        if f.is_file():
            return FileResponse(f.open(mode='rb'))
    else:
        return render(request, 'sunshine/cdn_403.html', context)


def cdn_serve(request, module, res_id, ext=''):
    ''' serve an uploaded content '''
    context = {
        'module': module,
        'res_id': res_id}
    if not request.user.is_authenticated:
        logging.getLogger(__name__).warning(f"CDN_BLOCKED: unauthorized access to {module} -- {res_id}")
        return render(request, 'sunshine/cdn_403.html', context)
        # return redirect(home)
    else:
        if module in SUNSHINE_CDN_PERMISSIONS:
            if request.user.groups.filter(name__in=SUNSHINE_CDN_PERMISSIONS[module]).exists():
                logging.getLogger(__name__).info(f"CDN_GRANTED: user {request.user} requested {module} -- {res_id}")
                f = MEDIA_ROOT / module / str(res_id)
                if f.exists():
                    return FileResponse(f.open(mode='rb'))
                else:
                    return render(request, 'sunshine/cdn_404.html', context)
            else:
                logging.getLogger(__name__).warning(f"CDN_BLOCKED: user {request.user} requested {module} -- {res_id}")
                return render(request, 'sunshine/cdn_403.html', context)
        else:
            return render(request, 'sunshine/cdn_403.html', context)
    return render(request, 'sunshine/cdn_403.html', context)



# -*- coding: utf-8 -*-

# This code is a part of sunshine package: https://github.com/letuananh/sunshine
# :copyright: (c) 2013-2021 Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

"""
Django command for listing all available permissions
"""

from django.core.management.base import BaseCommand
from django.contrib.auth.models import Permission


class Command(BaseCommand):
    help = 'List all current permissions'

    def handle(self, *args, **options):
        for p in Permission.objects.all():
            self.stdout.write(str(p))



# -*- coding: utf-8 -*-

# This code is a part of sunshine package: https://github.com/letuananh/sunshine
# :copyright: (c) 2013-2021 Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

import csv

from django import forms
from django.http import HttpResponse
from django.contrib import admin

from .common import read_settings


SUNSHINE_ENV = read_settings('SUNSHINE_ENV', 'DEV')
admin.site.site_header = read_settings('ADMIN_SITE_HEADER', f'{SUNSHINE_ENV} Sunshine Admin Portal')
admin.site.site_title = read_settings('ADMIN_SITE_TITLE', f'Sunshine-{SUNSHINE_ENV}')
admin.site.index_title = read_settings('ADMIN_SITE_INDEX_TITLE', 'Sunshine Portal administrator site')


class ExportCsvMixin:

    def export_as_csv(self, request, queryset):
        meta = self.model._meta
        field_names = [field.name for field in meta.fields]

        response = HttpResponse(content_type='text/csv')
        response['Content-Disposition'] = f'attachment; filename={meta}.csv'
        writer = csv.writer(response)

        writer.writerow(field_names)
        for obj in queryset:
            writer.writerow([getattr(obj, field) for field in field_names])
        return response

    export_as_csv.short_description = "Export selected"


class CsvImportForm(forms.Form):
    csv_file = forms.FileField()



# -*- coding: utf-8 -*-

# This code is a part of sunshine package: https://github.com/letuananh/sunshine
# :copyright: (c) 2020 Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

from datetime import datetime


def timestamp_now():
    ''' Get current timestamp in milliseconds '''
    return datetime.now().timestamp() * 1000


class CollabStatus:
    def __init__(self, current_user, last_edit):
        self.current_user = current_user
        self.last_edit = last_edit


class CollabManager:

    def __init__(self):
        self.__locks = {}
        self.__lock_users = {}

    def request_release(self, request):
        '''Release any pending lock by this user'''
        if request.user.username in self.__lock_users:
            return self.request_unlock(self.__lock_users[request.user.username], request)
        return True

    def request_lock(self, url, request):
        self.request_release(request)
        if request.user.username not in self.__lock_users:
            if url not in self.__locks:
                self.__locks[url] = request.user.username
                self.__lock_users[request.user.username] = url
            else:
                # already locked by someone else ...
                pass
        return url in self.__locks and self.__locks[url] == request.user.username

    def request_unlock(self, url, request):
        status = False
        if url in self.__locks and self.__locks[url] == request.user.username:
            status = self.__locks.pop(url) == request.user.username
            if request.user.username in self.__lock_users:
                try:
                    self.__lock_users.pop(request.user.username)
                except Exception:
                    pass
        return status

    def get_current_user(self, url):
        ''' Get user who is locking an url at the moment '''
        return self.__locks[url] if url in self.__locks else ''



# -*- coding: utf-8 -*-

# This code is a part of sunshine package: https://github.com/letuananh/sunshine
# :copyright: (c) 2013-2021 Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

import json
from django import template
import humanize


register = template.Library()


@register.filter('jsonify')
def jsonify(value):
    return json.dumps(value, ensure_ascii=False)


@register.filter('naturalsize')
def naturalsize(value):
    return humanize.naturalsize(value)



# -*- coding: utf-8 -*-

# This code is a part of sunshine package: https://github.com/letuananh/sunshine
# :copyright: (c) 2013-2021 Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

from django.urls import path
from . import views

urlpatterns = [
    path('', views.index, name="sunshine_index"),
    path('public/', views.public, name='sunshine_public'),
    path('home/', views.home, name='sunshine_home'),
    path('system/profile/', views.profile, name='sunshine_profile'),
    path('system/dashboard/', views.system_dashboard, name='system_dashboard'),
    path('logout/', views.logout, name='sunshine_logout'),
    # basic-CDN
    path('cdn/', views.cdn_home, name='sunshinecdn'),
    path(r'cdn/avatars/<slug:username>', views.cdn_serve_avatar, name='cdn_avatar'),
    path(r'cdn/direct/<slug:module>/<uuid:res_id>', views.cdn_serve, name='cdn_serve'),
    path(r'cdn/direct/<slug:module>/<uuid:res_id>/<slug:ext>', views.cdn_serve, name='cdn_serve'),
    # common APIs
    path('sunshine/api/common/sidebar',
         views.api_sidebar_toggle, name='ssapi_sidebar_toggle'),
    path('sunshine/api/common/sidebar/<int:status>',
         views.api_sidebar_toggle, name='ssapi_sidebar_toggle'),
    path('sunshine/api/notifications/fetch',
         views.api_fetch_notifications, name='ssapi_notification_fetch')
]
